dpstream.cu 4.3 KB

123456789101112131415161718192021222324252627282930313233343536373839404142434445464748495051525354555657585960616263646566676869707172737475767778798081828384858687888990919293949596979899100101102103104105106107108109110111112113114115116117118119120121122123124
  1. /*
  2. * cudadl version 0.9: Compute discrete logs in smooth group orders
  3. * using CUDA
  4. * Copyright (C) 2012 by Ryan Henry and Ian Goldberg
  5. * {rhenry,iang}@cs.uwaterloo.ca
  6. *
  7. * This program is free software: you can redistribute it and/or modify
  8. * it under the terms of version 3 of the GNU General Public License as
  9. * published by the Free Software Foundation.
  10. *
  11. * This program is distributed in the hope that it will be useful,
  12. * but WITHOUT ANY WARRANTY; without even the implied warranty of
  13. * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  14. * GNU General Public License for more details.
  15. *
  16. * You should have received a copy of the GNU General Public License
  17. * along with this program. If not, see <http://www.gnu.org/licenses/>.
  18. */
  19. #ifndef __DPSTREAM_H__
  20. #define __DPSTREAM_H__
  21. #include <sm_11_atomic_functions.h>
  22. /* Routines to allow CUDA threads to write distinguished points into
  23. a buffer that will be read by the host */
  24. __constant__ static unsigned int *DPbuffer = NULL;
  25. __constant__ static unsigned int DPbuffersize = 0; // in words
  26. static unsigned int *DPbuffer_device = NULL;
  27. __device__ volatile static unsigned int DPbuffertail = 0; // in words
  28. static const unsigned int DPrecordsize = (1+WORDS+6); // in words
  29. static unsigned int DPbufsize = 0;
  30. cudaError_t DPstreamInit(size_t numrecords)
  31. {
  32. // Allocate the buffer
  33. DPbufsize = DPrecordsize * numrecords;
  34. if (cudaMalloc((void **)&DPbuffer_device, DPbufsize * sizeof(unsigned int))
  35. != cudaSuccess) {
  36. return cudaErrorInitializationError;
  37. }
  38. // Zero it
  39. cudaMemset(DPbuffer_device, 0, DPbufsize * sizeof(unsigned int));
  40. // Initialize the device-visible pointers
  41. unsigned int zero = 0;
  42. cudaMemcpyToSymbol(DPbuffer, &DPbuffer_device, sizeof(unsigned int*));
  43. cudaMemcpyToSymbol(DPbuffersize, &DPbufsize, sizeof(unsigned int));
  44. cudaMemcpyToSymbol(DPbuffertail, &zero, sizeof(unsigned int));
  45. return cudaSuccess;
  46. }
  47. void DPstreamEnd(void)
  48. {
  49. if (DPbuffer_device) {
  50. cudaFree(DPbuffer_device);
  51. }
  52. DPbuffer_device = NULL;
  53. }
  54. __device__ inline void write_dp(unsigned int*, unsigned int, unsigned int,
  55. unsigned int, unsigned int, unsigned int, unsigned int);
  56. // Call this from the device to write the following into the buffer:
  57. // block and thread indices (1 word total)
  58. // x (WORDS words)
  59. // a (2 words)
  60. // b (2 words)
  61. __device__ inline void DPstreamWrite(unsigned int a_0, unsigned int a_1,
  62. unsigned int a_2, unsigned int b_0, unsigned int b_1, unsigned int b_2)
  63. {
  64. if (!DPbuffer) return;
  65. // Atomically increment the DPbuffertail
  66. if (DPbuffertail >= DPbuffersize) return;
  67. unsigned int ouroffset = atomicAdd((unsigned int *)&DPbuffertail,
  68. (unsigned int) (DPrecordsize));
  69. if (ouroffset + DPrecordsize > DPbuffersize) {
  70. atomicSub((unsigned int *)&DPbuffertail, (unsigned int)DPrecordsize);
  71. return;
  72. }
  73. unsigned int *ourbuffer = DPbuffer + ouroffset;
  74. write_dp(ourbuffer, a_0, a_1, a_2, b_0, b_1, b_2);
  75. }
  76. static void dpcallback(void *data, unsigned short threadId,
  77. unsigned short blockId, string x, unsigned int a_0, unsigned int a_1,
  78. unsigned int a_2, unsigned int b_0, unsigned int b_1, unsigned int b_2);
  79. void DPstreamParse(void *data)
  80. {
  81. // Get the number of words in the buffer
  82. unsigned int bufsize;
  83. cudaMemcpyFromSymbol(&bufsize, DPbuffertail, sizeof(unsigned int));
  84. cerr << getpid() << " " << bufsize / DPrecordsize << " DPs\n";
  85. unsigned int *dpbuf = (unsigned int*)calloc(bufsize, sizeof(unsigned int));
  86. if (dpbuf) {
  87. cudaMemcpy(dpbuf, DPbuffer_device, bufsize * sizeof(unsigned int),
  88. cudaMemcpyDeviceToHost);
  89. for (unsigned int *dp=dpbuf; dp < dpbuf+bufsize; dp += DPrecordsize) {
  90. unsigned short threadId = (unsigned short)(dp[0]);
  91. unsigned short blockId = dp[0]>>16;
  92. unsigned int a_0 = dp[WORDS+1];
  93. unsigned int a_1 = dp[WORDS+2];
  94. unsigned int a_2 = dp[WORDS+3];
  95. unsigned int b_0 = dp[WORDS+4];
  96. unsigned int b_1 = dp[WORDS+5];
  97. unsigned int b_2 = dp[WORDS+6];
  98. string x((const char *)(dp+1), WORDS*sizeof(unsigned int));
  99. dpcallback(data, threadId, blockId, x, a_0, a_1, a_2, b_0, b_1, b_2);
  100. }
  101. }
  102. free(dpbuf);
  103. // Reset the buffer
  104. unsigned int zero = 0;
  105. cudaMemcpyToSymbol(DPbuffertail, &zero, sizeof(unsigned int));
  106. }
  107. #endif